/**
 * Layer to encapsulate the keyword extraction API and
 * make it accessible to gnunet-insert.
 * @file src/textui/keywords.c
 * @author Christian Grothoff
 **/

#include "textui/keywords.h"
#include "util/xmalloc.h"
#include "configuration.h"
#include <string.h>
#ifdef USE_EXTRACTOR
#include <extractor.h>
#endif

/**
 * Extract keywords, mime-type and description from a file
 * @param filename the name of the file
 * @param description the description (the user may have
 *        supplied a description already (*description != NULL),
 *        in that case, append, mind the maximum size!
 * @param mimetype the mimetype, again, the user may
 *        have supplied one
 * @param keywords the list of keywords, allocate space at
 *        another location if required, copy existing keywords
 *        over to that space! Do NEVER free *keywords!
 * @param num_keywords the number of keywords in the
 *        existing *keywords array that was passed in.
 *        Set *num_keywords to the new number of keywords!
 **/
void extractKeywords(char * filename,
		     char ** description,
		     char ** mimetype,
		     char *** keywords,
		     int * num_keywords) {
#ifdef USE_EXTRACTOR
  EXTRACTOR_ExtractorList * exList;
  exList = getExtractors();
  extractKeywordsMulti(filename, description,
		       mimetype, keywords, 
		       num_keywords, exList);
  EXTRACTOR_removeAll(exList);
#else
  *num_keywords = 0;
#endif
}

#ifdef USE_EXTRACTOR
EXTRACTOR_ExtractorList * getExtractors() {
  char * config;
  EXTRACTOR_ExtractorList * exList;

  exList = EXTRACTOR_loadDefaultLibraries();
  config = getExtractorConfig();
  exList = EXTRACTOR_loadConfigLibraries(exList,
				       config);
  xfree(config, "extractor configuration (strdup)");
  return exList;
}
#endif

/**
 * Extract keywords, mime-type and description from a file
 * @param filename the name of the file
 * @param description the description (the user may have
 *        supplied a description already (*description != NULL),
 *        in that case, append, mind the maximum size!
 * @param mimetype the mimetype, again, the user may
 *        have supplied one
 * @param keywords the list of keywords, allocate space at
 *        another location if required, copy existing keywords
 *        over to that space! Do NEVER free *keywords!
 * @param num_keywords the number of keywords in the
 *        existing *keywords array that was passed in.
 *        Set *num_keywords to the new number of keywords!
 **/
void extractKeywordsMulti(char * filename,
			  char ** description,
			  char ** mimetype,
			  char *** keywords,
			  int * num_keywords,
			  EXTRACTOR_ExtractorList * exList) {
#ifdef USE_EXTRACTOR
  EXTRACTOR_KeywordList * keyList;
  EXTRACTOR_KeywordList * pos;
  char ** newKeywords;
  char * key;
  int count;
  int i;
  keyList = EXTRACTOR_getKeywords(exList, filename);
  keyList = EXTRACTOR_removeDuplicateKeywords(keyList,
					      EXTRACTOR_DUPLICATES_REMOVE_UNKNOWN);
  if (*mimetype == NULL) {
    key = EXTRACTOR_extractLast(EXTRACTOR_MIMETYPE, keyList);
    if (key != NULL)
      *mimetype = strdup(key);
  }
  if (*description == NULL) {
    key = EXTRACTOR_extractLast(EXTRACTOR_DESCRIPTION, keyList);
    if (key != NULL)    
      *description = strdup(key);
  }
  keyList = EXTRACTOR_removeDuplicateKeywords(keyList,
					      EXTRACTOR_DUPLICATES_TYPELESS);
  count = EXTRACTOR_countKeywords(keyList);
  newKeywords = (char**) xmalloc((count+(*num_keywords)) * sizeof(char*),
				 "extractKeywords: new keywords");
  for (i=0;i<*num_keywords;i++)
    newKeywords[i] = (*keywords)[i];
  pos = keyList;
  for (i=0;i<count;i++) {
    newKeywords[*num_keywords+i] = strdup(pos->keyword);
    pos = pos->next;
  }
  /* assert(keyList == NULL); */
  EXTRACTOR_freeKeywords(keyList);
  *keywords = newKeywords;
  *num_keywords = *num_keywords + count;
#else
  *num_keywords = 0;
#endif
}
